/*
 * test_mem.c
 *      to test the memory allocation routines
 * 1993/08  by kmatsui
 * 1996/09  by kmatsui
 *      Added t_sbrk(), t_write_free() and t_corr_block().
 *      Added routines to test several packages of malloc().
 *      Revised some trivials.
 * 1997/03  by kmatsui
 *      Added call to set_memlist().
 * 1998/05  by kmatsui
 *      Added implementation for Win32 and phk-malloc.
 * 1998/11  by kmatsui
 *      Updated following revision of set_memfp().
 *      Added test for DLMALLOC.
 * 2003/06  by kmatsui
 *      Slightly modified output fp.
 * 2003/09  by kmatsui
 *      Slightly modified output fp again.
 * 2006/01  by kmatsui
 *      Added t_m0_f();
 *
 *  One of the following 9 macros may be defined by command line.
 *  KMMALLOC    :   kmatsui's version of malloc().
 *  BSD_MALLOC  :   for the malloc() of 4.[34]BSD (and GNU C / libc)
 *                  compiled with -DMSTATS (maybe also -DDEBUG -DRCHECK).
 *  PHK_MALLOC  :   for the malloc() of Poul-Henning Kamp (and FreeBSD)
 *                  compiled with -DMALLOC_EXTRA_SANITY and -DMALLOC_STATS.
 *  DLMALLOC    :   for the malloc() of Doug Lea.
 *  DB_MALLOC   :   for the debugging malloc() by Conor P. Cahill.
 *  DMALLOC     :   for the debug malloc() by Gray Watson.
 *  LEAK        :   for the leak by Ch. G. Phillips.
 *  BORLANDC    :   for the malloc() of Borland C.
 *  VISUALC     :   for the malloc() of Visual C++.
 *  none of the above 9 macros  :   for any malloc() without listing blocks.
 *
 *  LARGE       :   MS-DOS large data memory model.
 */

#include    <stdio.h>
#include    <string.h>
#include    <errno.h>
#if     ! DMALLOC               /* To avoid declaration mismatch    */
#include    <stdlib.h>
#endif

#if     KMMALLOC
#include    "yalloc.h"
extern FILE *memfp;
#elif   LARGE
    typedef unsigned long   _ptr_t;
#else
    typedef unsigned        _ptr_t;
#if     DB_MALLOC
#include    "malloc.h"
#elif   DMALLOC
#include    "dmalloc.h"
#elif   LEAK
#include    "leak.h"
#endif
#endif

#define     _PMAX           (~(_ptr_t)0)
#define     ULONG( val)     ((long) (val) & _PMAX)

#if     defined __WIN32 || defined _WIN32 || defined WIN32
#define WIN32API    1
#endif

#if     __GO32__
#define PLUS        0x400000    /* Address of data area */
#else
#define PLUS        0
#endif

FILE    *fp;
char    dummy[ 4096] = "static array of 4096 bytes";
char    *alloc = "allocated area";
char    *nulstr = "NULL";

#if     PROTO
static void     mem_init( int, char **);
static void     t_brk( void);
static void     t_sbrk( int);
static void     t_m0_f( void);
static void     t_m_f( void);
static void     t_c_f( void);
static void     t_m_r( void);
static void     t_ill_free( void);
static void     t_write_free( void);
static void     t_corr_blk( void);
static void     prt_malloc( char *, size_t);
static void     prt_free( char *);
static void     list( void);
#else
static void     mem_init();
static void     t_brk();
static void     t_sbrk();
static void     t_m0_f();
static void     t_m_f();
static void     t_c_f();
static void     t_m_r();
static void     t_ill_free();
static void     t_write_free();
static void     t_corr_blk();
static void     prt_malloc();
static void     prt_free();
static void     list();
#endif

#if     KMMALLOC
#if     PROTO
extern int      list_heap( int);
extern void     print_memerror( int);
extern void     memerr_list( int);
#else
extern int      list_heap();
extern void     print_memerror();
extern void     memerr_list();
#endif
#else
#if     PROTO
extern int      list_heap( FILE *);
#else
extern int      list_heap();
#endif
#endif

main( argc, argv)
    int     argc;
    char    **argv;
{
    fp = stderr;            /* Not to buffer        */
    list();                 /* Already any heap ?   */
    mem_init( argc, argv);
#if     __unix__
    {   /* Show the size of code and static data    */
        extern const void   etext, edata, end;
        fprintf( fp, "etext:0x%x, edata:0x%x, end:0x%x\n",
                &etext, &edata, &end);
    }
#endif
    list();                 /* Already any heap ?   */

#if     T_BRK
    t_brk();
#endif
#if     T_SBRK
    t_sbrk( 0x12);
#endif

    t_m0_f();
    list();

    t_m_f();
    list();

    t_c_f();
    list();

#if     T_SBRK
    t_sbrk( 0x123);
#endif

    t_m_r();
    list();

#if     ! __unix__ && ! WIN32API
        /* Probably cause segmantation fault    */
    t_ill_free();
    list();
#endif

#if     KMMALLOC
    merrfunc = (void(*)()) 0;       /* Do not execute any function  */
#endif
    t_write_free();
    list();

#if     KMMALLOC
    merrfunc = memerr_list;     /* Yet another function on error    */
#endif
    t_corr_blk();
    list();

#if     LEAK
    leak_close();   /* Flush data-base on memory to the corresponding files */
#endif

    return  errno;
}

static void     mem_init( argc, argv)
    int     argc;
    char    **argv;
{
#if     KMMALLOC

    extern int  set_memcheck( void);        /* Getenv and set 'memcheck'    */
    extern unsigned set_memlist( void);     /* Getenv and set 'memlist'     */
    extern void (*set_merrfunc(void))();    /* Getenv and set 'merrfunc'    */

    if (set_memcheck() == -1)
        memcheck |= MERR_CHKALL;
    if (set_merrfunc() == (void (*)())NULL)
        merrfunc = print_memerror;
    if (memfp == stdout || memfp == NULL)
        memfp = fp;                 /* stderr   */
    else
        fp = memfp;                 /* "memlist" ?  */

#elif   DB_MALLOC

    union dbmalloptarg      m;
    m.i = M_HANDLE_IGNORE;
    dbmallopt( MALLOC_WARN, &m);
    dbmallopt( MALLOC_FATAL, &m);
    m.i = 1;
    dbmallopt( MALLOC_CKCHAIN, &m);

#elif   DMALLOC

    /* Options are set reading environment variable by malloc routines  */

#endif

#if     (WIN32API || __MSDOS__ || LSI_C) && !(KMMALLOC && NO_STDOUT)
                                    /* Redirect 'fp' (possibly stderr)  */
    freopen( "test_mem.out", "w", fp);
#endif

    setbuf( fp, NULL);
}

#if     T_BRK
static void     t_brk()
{
#if     PROTO
    extern int  brk( void *);
#else
    extern int  brk();
#endif
    int         brk_value;
    char        *ptr;

    for (ptr = (char *)0xc000 + PLUS; ptr > (char *)0x8000 + PLUS;
            ptr -= 0x1234) {
        errno = 0;
        brk_value = brk( ptr);
        fprintf( fp, "brk(0x%lx) -- value:%d, errno:%d\n"
                , ULONG( ptr), brk_value, errno);
    }
}
#endif

#if     T_SBRK
static void     t_sbrk( size)
    int     size;
{
#if     PROTO
    extern void *sbrk( int);
#else
    extern char *sbrk();
#endif
    char    *ptr1, *ptr2;

    ptr1 = sbrk( size);
    ptr2 = sbrk( 0);
    fprintf( fp, "0x%lx = sbrk(0x%x) -> 0x%lx\n",
            ULONG( ptr1), size, ULONG( ptr2));
}
#endif

static void     t_m0_f()
/* sample to test handling of 0-sized block -- provided by freebeans    */
{
    char    *ptr1, *ptr2, *ptr3;

    ptr1 = malloc( 100);
    prt_malloc( ptr1, 100);
    ptr2 = malloc( 0);
    prt_malloc( ptr2, 0);   /* 0-sized block sandwiched by non0-sized blocks*/
    ptr3 = malloc( 100);
    prt_malloc( ptr3, 100);
    free( ptr2);
    prt_free( ptr2);
    free( ptr3);
    prt_free( ptr3);
    free( ptr1);
    prt_free( ptr1);
}

static void     t_m_f()
{
    char    *ptr;
    size_t  malloc_max, u;
    int     i;

    if (sizeof (size_t) > 2)
        malloc_max = 0x1fb00;
    else
        malloc_max = 0xfb00;
    fprintf( fp, "dummy:0x%lx -> %s\n", ULONG( dummy), dummy);
    for (i = 0, u = malloc_max; i < 5; i++, u -= 0x3210) {
        errno = 0;
        ptr = (char *) malloc( u);
        prt_malloc( ptr, u);
        if (ptr) {
            errno = 0;
            free( ptr);
            prt_free( ptr);
        }
    }
}

static void     t_c_f()
{
    char    *ptr;
    size_t  u, u2;

    for (u = 0x12, u2 = 0x23; u < 0xc0; u += 0x34, u2 += 0x45) {
        errno = 0;
        ptr = calloc( u, u2);
        if (ptr)
            strcpy( ptr, alloc);
        fprintf( fp, "calloc(0x%lx,0x%lx) -- errno:%d, ptr:0x%lx -> %s\n"
                , ULONG( u), ULONG( u2), errno
                , ULONG( ptr), ptr ? ptr : nulstr);
        if (ptr) {
            errno = 0;
            free( ptr);
            prt_free( ptr);
            errno = 0;
            free( ptr);                         /* Freeing twice    */
            prt_free( ptr);
        }
    }
}

static void     t_m_r()
{
    char    *ptr, *ptr2;
    size_t  u, u2;

    for (u = 0x8000, u2 = 0; u <= 0x8000; u -= 0x1357, u2 += 0x1357) {
        errno = 0;
        ptr = (char *) malloc( u);
        if (ptr)
            strcpy( ptr, alloc);
        fprintf( fp, "malloc(0x%lx) -- errno:%d, ptr:0x%lx -> %s\n"
                , ULONG( u), errno, ULONG( ptr), ptr ? ptr : nulstr);
        errno = 0;
        ptr2 = (char *) realloc( ptr, u2);
        fprintf( fp, "realloc(0x%lx,0x%lx) -- errno:%d, ptr2:0x%lx -> %s\n"
                , ULONG( ptr), ULONG( u2), errno
                , ULONG( ptr2), (ptr && ptr2 && u2) ? ptr2 : nulstr);
    }
}

static void     t_ill_free()
{
    char    *ptr;

    for (ptr = (char *)0x1000 + PLUS;
            ptr >= (char *)0x1000 + PLUS && ptr < (char *)0xe000 + PLUS;
            ptr += 0x3456) {
        errno = 0;
        free( ptr);                             /* Illegal pointers */
        prt_free( ptr);
    }
}

static void     t_write_free()
{
    char    *ptr;
    size_t  u;

    errno = 0;
    u = 12;
    ptr = (char *) malloc( u);
    prt_malloc( ptr, u);
    if (ptr) {
        errno = 0;
        free( ptr);
        prt_free( ptr);
        strcpy( ptr, "botch");              /* Writing on freed area    */
    }
    errno = 0;
    u = 9;
    ptr = (char *) malloc( u);
    prt_malloc( ptr, u);
    if (ptr) {
        strcpy( ptr, "123456789");  /* Typical bug of overrunning block */
        errno = 0;
        free( ptr);
        prt_free( ptr);
    }
}

static void     t_corr_blk()
{
    char    *ptr;
    size_t  u;

    errno = 0;
    u = 0x123;
    ptr = malloc( u);
    prt_malloc( ptr, u);
    *(ptr - 1) = '\xFF';                    /* Corrupt the header   */
    free( ptr);
    prt_free( ptr);
    errno = 0;
    u = 0x234;
    ptr = malloc( u);
    prt_malloc( ptr, u);
    errno = 0;
    strcpy( ptr + u + 1, "botch botch");
                                /* Perhaps corrupt the next header  */
    u = 0x345;
    ptr = malloc( u);
    prt_malloc( ptr, u);
}

static void     prt_malloc( ptr, u)
    char    *ptr;
    size_t  u;
{
    fprintf( fp, "malloc(0x%lx) -- errno:%d, ptr:0x%lx\n"
            , ULONG( u), errno, ULONG( ptr));
}

static void     prt_free( ptr)
    char    *ptr;
{
#if     KMMALLOC
    _Header     *header;

    header = (_Header *) (ptr - _CELLSIZE);
    fprintf( fp, "free(0x%lx) -- errno:%d", ULONG( ptr), errno);
    if (errno != 0 || ptr == NULL)
        fprintf( fp, "\n");
    else
        fprintf( fp, ", header:0x%lx, _size:0x%lx, _next:0x%lx\n",
                ULONG( header), ULONG( header->_size), ULONG( header->_next));
#else
    fprintf( fp, "free(0x%lx) -- errno:%d\n", ULONG( ptr), errno);
#endif
}

static void     list()
{
    fflush( fp);
#if     KMMALLOC
    list_heap( 1);
#elif   1                           /* 0: not to call list_heap()   */
#if     BSD_MALLOC
    list_heap( ": test_mem");
#elif   VISUALC || BORLANDC || PHK_MALLOC || DLMALLOC || DB_MALLOC  \
        || DMALLOC || LEAK
    list_heap( fp);
#endif
#endif
    fflush( fp);

#if     LEAK
    fflush( stdout);                /* leak_dump() uses stdout      */
#endif
}

