/*
 * time_mem.c
 *      to measure the speed of memory allocation routines
 * 1993/08      kmatsui
 * 1996/09      kmatsui
 *      Substituted time() with clock().
 *      Added routines to test some packages of malloc().
 *      Revised m_r().
 * 1998/05      kmatsui
 *      Added implementation for Win32 and phk-malloc.
 * 1998/11      kmatsui
 *      Updated following the revision of set_memfp().
 *      Added test for DLMALLOC.
 * 2003/09      kmatsui
 *      Modified output fp.
 *
 *  One of the following 9 macros may be defined by command line.
 *  KMMALLOC    :   kmatsui's version of malloc().
 *  BSD_MALLOC  :   for the malloc() of 4.[34]BSD (and GNU C / libc)
 *                  compiled with -DMSTATS (maybe also -DDEBUG -DRCHECK).
 *  PHK_MALLOC  :   for the malloc() of Poul-Henning Kamp (and FreeBSD)
 *                  compiled with -DMALLOC_EXTRA_SANITY and -DMALLOC_STATS.
 *  DLMALLOC    :   for the malloc() of Doug Lea.
 *  DB_MALLOC   :   for the debugging malloc() by Conor P. Cahill.
 *  DMALLOC     :   for the debug malloc() by Gray Watson.
 *  LEAK        :   for the leak by Ch. G. Phillips.
 *  BORLANDC    :   for the malloc() of Borland C.
 *  VISUALC     :   for the malloc() of Visual C++.
 *  none of the above 9 macros  :   for any malloc() without listing blocks.
 *
 *  LARGE       :   MS-DOS large data memory model.
 */

#include    <stdio.h>
#include    <time.h>
#include    <errno.h>
#if     ! DMALLOC
#include    <stdlib.h>
#endif

#if     KMMALLOC
#include    "xalloc.h"
#elif   DB_MALLOC
#include    "malloc.h"
#elif   DMALLOC
#include    "dmalloc.h"
#elif   LEAK
#include    "leak.h"
#endif

#if     PROTO
static void     mem_init( int, char **);
static void     m_f( void);
static void     m_r_r( int);
static void     m_r_f( int);
static void     list( void);
#if     KMMALLOC
extern int      list_heap( int);
#elif   BSD_MALLOC
extern int      list_heap( char *);
#else
extern int      list_heap( FILE *);
#endif
#else
static void     mem_init();
static void     m_f();
static void     m_r_r();
static void     m_r_f();
static void     list();
extern int      list_heap();
#endif

#if     (__MSDOS__ || LSI_C) && !__GO32__
#define ITER    100 /* clock() returns time in units of a second    */
#else
#define ITER    10
#endif

#if     LARGE
#define SIZE    512
#define ASIZE   512
#elif   ((__MSDOS__ || LSI_C) && !__GO32__) || __MWC09__
#define SIZE    256
#define ASIZE   256
#else
#define SIZE   4096
#define ASIZE   512
#endif

FILE    *fp;
unsigned    ms, rs, fs;
unsigned    u, u2;
long    r_size, a_size;
char    *hptr[ ASIZE];

#if     KMMALLOC
    extern FILE     *memfp;
#endif

main( argc, argv)
    int     argc;
    char    **argv;
{
    int     iter;
    char    *ticks = "%ld ticks passed.\n";
    long    t_start, t_cur;

    mem_init( argc, argv);
    t_start = clock();

    for (iter = 0; iter < ITER; iter++)
        m_f();

    t_cur = clock();
    fprintf( fp, ticks, t_cur - t_start);
    fprintf( fp, "Executed malloc() and free() %u times.\n", ms * ITER);
    list();

    for (iter = 0; iter < ITER; iter++)
        m_r_r( iter);

    t_cur = clock();
    fprintf( fp, ticks, t_cur - t_start);
    fprintf( fp, "Executed malloc() %u times and realloc() %u times.\n",
            ms * ITER, rs * ITER);
    fprintf( fp, "Total size of %ld bytes has been re-allocated.\n",
            r_size * ITER);
    list();

    for (iter = 0; iter < ITER; iter++) {
        m_r_f( iter);
        if (iter == 0 || iter == ITER - 1) {
#if ITER > 1
            fprintf( fp, "iter %d : ", iter);
#endif
            fprintf( fp,
            "Total %ld bytes (net) of %u blocks are allocated at present.\n",
                        a_size, ms - fs);
        }
    }

    t_cur = clock();
    fprintf( fp, ticks, t_cur - t_start);
    fprintf( fp,
        "Executed malloc() %u times, realloc() %u times, free %u times.\n",
            ms * ITER, rs * ITER, fs * ITER);
    fprintf( fp, "Total size of %ld bytes has been re-allocated.\n",
            r_size * ITER);
    list();

#if     DMALLOC
    dmalloc_shutdown();
#elif   LEAK
    leak_close();   /* Flush data-base on memory to the corresponding files */
#endif
    return  errno;
}

static void     mem_init( argc, argv)
    int     argc;
    char    **argv;
{
    fp = stdout;
#if     KMMALLOC

    extern int  set_memcheck( void);        /* Getenv and set 'memcheck'    */
    extern FILE *set_memfp( void);          /* Getenv and set 'memfp'       */
    extern void (*set_merrfunc( void))();   /* Getenv and set 'merrfunc'    */

    set_memcheck();
    set_merrfunc();
    set_memfp();
    list();                                 /* already any heap?    */
    if (memfp == NULL)                      /* no heap yet          */
        memfp = fp;
    else
        fp = memfp;

#elif   DB_MALLOC

    union dbmalloptarg  m;
    m.i = 0;                            /* disable error checking   */
#if     1
    dbmallopt( MALLOC_WARN, &m);
    dbmallopt( MALLOC_FATAL, &m);
    dbmallopt( MALLOC_CKCHAIN, &m);
    dbmallopt( MALLOC_CKDATA, &m);
#endif
    dbmallopt( MALLOC_FILLAREA, &m);

#elif   DMALLOC

    extern char     *getenv( const char *);
    if (getenv( "DMALLOC_OPTIONS") == NULL)
        dmalloc_debug( 0L);             /* disable error checking   */

#elif   LEAK

    leak_logging = 1;

#endif
}

static void m_f()
{
    char    *ptr;

    for (ms = 0, u = 1, ptr = (char *)1; u <= 0xC000 && ptr != NULL; u += 8) {
        ptr = (char *) malloc( u);
        free( ptr);
        ms++;
    }
}

static void m_r_r( iter)
    int     iter;
{
    int     i;
    char    *ptr;

    for (r_size = 0, rs = 0, ms = 0; ms < ASIZE; ms++) {
        if ((ptr = (char *) malloc( SIZE * 16)) == NULL)
            break;
        else
            hptr[ ms] = ptr;
        if ((ptr = (char *) realloc( ptr, SIZE / 2)) == NULL)
            break;
        else {
            hptr[ ms] = ptr;
            r_size += SIZE / 2;
            rs++;
        }
        if ((ptr = (char *) realloc( ptr, (SIZE * 3) / 4)) == NULL)
            break;
        else {
            hptr[ ms] = ptr;
            r_size += (SIZE * 3) / 4;
            rs++;
        }
    }
    if (iter == ITER - 1)
        list();
    if (ms < ASIZE)
        ms++;
    for (i = 0; i < ms; i++)
        free( hptr[ i]);
}

static void m_r_f( iter)
    int     iter;
{
    size_t  fsize[ 4];
    int     unit1, unit2, i;
    char    *ptr;

    if (sizeof (size_t) >= 4) {
        u2 = 0x60E0;
        unit1 = 0x20;
        unit2 = 0x40;
    } else {
#if     LARGE
        unit1 = 8;
#if     __TURBOC__ >= 0x0452
        u2 = 0x12A0;
        unit2 = 24;
#else
        u2 = 0x13E8;
        unit2 = 25;
#endif
#else
#if     __TURBOC__ >= 0x0452
        u2 = 0x4B4;
#else
        u2 = 0x4C8;
#endif
        unit1 = 2;
        unit2 = 12;
#endif
    }
    r_size = a_size = ms = rs = fs = 0;

    for (u = 0x200 * unit1;
            ms < ASIZE && u && u2 && u <= 0x7FFF && u2 <= 0x7FFF;
            u -= unit1, u2 -= unit2) {
        if ((ptr = (char *) malloc( u)) == NULL)
            break;
        hptr[ ms++] = ptr;
        if ((ptr = (char *) realloc( ptr, u2)) == NULL) {
            if (u2)
                a_size += u;
            break;
        }
        hptr[ rs++] = ptr;
        r_size += u2;
        a_size += u2;
        if ((rs & 3) == 1) {
            i = (rs >> 2) & 3;
            fsize[ i] = u2;
        } else if ((rs & 15) == 0) {
            for (i = 0; i < 4; i++) {
                free( hptr[ rs + i * 4 - 16]);
                hptr[ rs + i * 4 - 16] = NULL;
                a_size -= fsize[ i];
            }
            fs += 4;
            if (--unit2 < 0)
                break;
        }
    }
    if (iter == 0 || iter == ITER - 1)
        fprintf( fp, "errno:%d, unit2:%d, u:0x%04x, u2:0x%04x\n",
                errno, unit2, u, u2);
    if (iter == ITER - 1)
        list();
    for (i = 0; i < ms; i++)
        free( hptr[ i]);
}

static void     list()
{
#if     KMMALLOC
    list_heap( 1);
#elif   1       /* 0: not to link list_heap() for the following malloc()s   */
#if     BSD_MALLOC
    list_heap( ": time_mem");
#elif   VISUALC || BORLANDC || PHK_MALLOC || DLMALLOC || DB_MALLOC  \
            || DMALLOC || LEAK
    list_heap( stdout);
#endif
#endif
}

